from __future__ import absolute_import
import wx
from wx.lib.pubsub import Publisher
from wx.lib.pubsub.core.datamsg import Message
import matplotlib
matplotlib.interactive(True)
matplotlib.use('WXAgg')
from matplotlib.figure import Figure
try:
    from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg
except IOError:
    # on some linux installations this import needs to be done twice as the first time raises an error:
    #   IOError: [Errno 2] No such file or directory: '/tmp/matplotlib-parallels/fontList.cache'
    from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg
from matplotlib.patches import PathPatch
from matplotlib.path import Path
import numpy as np
import sys


class PlotPlotPanel(wx.Panel):
    def __init__(self, parent, dpi=None, **kwargs):
        wx.Panel.__init__(self, parent, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, **kwargs)
        self.ztv_frame = self.GetTopLevelParent()
        self.figure = Figure(dpi=None, figsize=(1.,1.))
        self.axes = self.figure.add_subplot(111)
        self.canvas = FigureCanvasWxAgg(self, -1, self.figure)
        self.Bind(wx.EVT_SIZE, self._onSize)

    def _onSize(self, event):
        self._SetSize()

    def _SetSize(self):
        pixels = tuple(self.GetClientSize())
        self.SetSize(pixels)
        self.canvas.SetSize(pixels)
        self.figure.set_size_inches(float(pixels[0])/self.figure.get_dpi(), float(pixels[1])/self.figure.get_dpi())



class PlotPanel(wx.Panel):
    def __init__(self, parent):
        wx.Panel.__init__(self, parent, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.TAB_TRAVERSAL)
        self.ztv_frame = self.GetTopLevelParent()
        self.primary_image_patch = None

        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.plot_panel = PlotPlotPanel(self)
        self.sizer.Add(self.plot_panel, 1, wx.LEFT | wx.TOP | wx.EXPAND)
        
        self.hideshow_button = wx.Button(self, wx.ID_ANY, u"Hide", wx.DefaultPosition, wx.DefaultSize, 0)
        self.sizer.Add(self.hideshow_button, 0, wx.ALL|wx.ALIGN_RIGHT|wx.ALIGN_BOTTOM, 2)
        self.hideshow_button.Bind(wx.EVT_BUTTON, self.on_hideshow_button)

        self.SetSizer(self.sizer)
        self.Fit()
        self.start_pt = wx.RealPoint(0., 0.)
        self.end_pt = wx.RealPoint(0., 0.)
        self.redraw()
        Publisher().subscribe(self.on_new_xy0, "new_slice_plot_xy0")
        Publisher().subscribe(self.on_new_xy1, "new_slice_plot_xy1")
        Publisher().subscribe(self.redraw, "primary_xy_limits-changed")
        Publisher().subscribe(self.redraw, "redraw_image")

    def on_new_xy0(self, msg):
        if isinstance(msg, Message):
            x,y = msg.data
        else:
            x,y = msg
        self.start_pt.x, self.start_pt.y = x, y
        self.redraw_overplot_on_image()
        self.redraw()

    def on_new_xy1(self, msg):   
        if isinstance(msg, Message):
            x,y = msg.data
        else:
            x,y = msg
        self.end_pt.x, self.end_pt.y = x, y
        self.redraw_overplot_on_image()
        self.redraw()

    def redraw_overplot_on_image(self):
        if self.primary_image_patch is not None:
            self.ztv_frame.primary_image_panel.axes.patches.remove(self.primary_image_patch)
        path = Path([self.start_pt, self.end_pt], [Path.MOVETO, Path.LINETO])
        self.primary_image_patch = PathPatch(path, color='magenta', lw=1)
        self.ztv_frame.primary_image_panel.axes.add_patch(self.primary_image_patch)
        self.ztv_frame.primary_image_panel.figure.canvas.draw()
        self.hideshow_button.SetLabel(u"Hide")        

    def remove_overplot_on_image(self):
        if self.primary_image_patch is not None:
            self.ztv_frame.primary_image_panel.axes.patches.remove(self.primary_image_patch)
        self.ztv_frame.primary_image_panel.figure.canvas.draw()
        self.primary_image_patch = None
        self.hideshow_button.SetLabel(u"Show")

    def on_hideshow_button(self, event):
        if self.hideshow_button.GetLabel() == 'Hide':
            self.hideshow_button.SetLabel(u"Show")
            self.remove_overplot_on_image()
        else:
            self.hideshow_button.SetLabel(u"Hide")        
            self.redraw_overplot_on_image()
            
    def redraw(self, *args):
        xy_limits = self.ztv_frame.primary_image_panel.set_and_get_xy_limits()
        oversample_factor = 10.
        n_pts = oversample_factor*np.max(self.ztv_frame.display_image.shape)
        xs = np.linspace(self.start_pt.x, self.end_pt.x, n_pts)
        ys = np.linspace(self.start_pt.y, self.end_pt.y, n_pts)
        mask = ((xs >= min(xy_limits['xlim'])) & (xs <= max(xy_limits['xlim'])) & 
                (ys >= min(xy_limits['ylim'])) & (ys <= max(xy_limits['ylim'])))
        xs = xs[mask]
        ys = ys[mask]
        if len(xs) > 0:
            if (ys.max() - ys.min()) > (xs.max() - xs.min()):   # dominantly a vertical slice
                if ys[0] > ys[1]:
                    xs = xs[-1::-1]
                    ys = ys[-1::-1]
            else:
                if xs[0] > xs[1]:
                    xs = xs[-1::-1]
                    ys = ys[-1::-1]
            if np.min(np.round(xs)) == np.max(np.round(xs)):
                positions = ys
            elif np.min(np.round(ys)) == np.max(np.round(ys)):
                positions = xs
            else:
                positions = np.sqrt( (xs - xs[0])**2 + (ys - ys[0])**2 )
            xs = xs.astype(np.int)
            ys = ys.astype(np.int)
            im_values = self.ztv_frame.display_image[ys, xs]
            self.plot_panel.axes.clear()
            if positions.min() != positions.max():
                self.line_plot = self.plot_panel.axes.plot(positions, im_values)
                self.plot_panel.axes.set_xlim([positions[0], positions[-1]])
            self.plot_panel.figure.canvas.draw()
        else:
            self.plot_panel.axes.clear()
            self.plot_panel.figure.canvas.draw()
        


import wx
import sys
import pickle
from threading import Thread
from Queue import Queue, Empty

def force_textctrl_color_update(textctrl):
    cur_focused_item = textctrl.GetParent().FindFocus()
    insertion_point = textctrl.GetInsertionPoint()
    children = textctrl.GetParent().GetChildren()
    can_accept_focus_mask = [(hasattr(a, 'CanAcceptFocus') and a.CanAcceptFocus() and 
                              (a is not textctrl)) for a in children]
    if True in can_accept_focus_mask:
        children[can_accept_focus_mask.index(True)].SetFocus()
    textctrl.SetFocus()
    textctrl.SetInsertionPoint(insertion_point)
    if cur_focused_item is not None:
        cur_focused_item.SetFocus()

def set_textctrl_background_color(textctrl, mode, tooltip=None):
    if mode == 'ok':
        color = (255,255,255)
    elif mode == 'enter-needed':
        color = (200,255,200)
    elif mode == 'invalid':
        # TODO:  implement: escape key brings up last valid value??
        color = (255,200,200)
    textctrl.SetBackgroundColour(color)
    textctrl.Refresh()
    if tooltip is not None and not isinstance(tooltip, wx.ToolTip):
        tooltip = wx.ToolTip(tooltip)
    textctrl.SetToolTip(tooltip)
    force_textctrl_color_update(textctrl)

def validate_textctrl_str(textctrl, validate_fxn, last_value):
    """
    can accept arbitrary functions in validate_fxn.  They just need to raise a ValueError if
    they don't like the input.
    """
    try:
        newval = validate_fxn(textctrl.GetValue())
        if textctrl.GetValue() == last_value:
            set_textctrl_background_color(textctrl, 'ok')
        else:
            set_textctrl_background_color(textctrl, 'enter-needed',
                                         'Press enter in this field to set new minimum value')
        return True
    except ValueError:
        # TODO: figure out some (clever?) way of having validate_fxn give info about itself that is more useful in the following error tooltip message
        set_textctrl_background_color(textctrl, 'invalid', 
                                      'Entry cannot be converted to {}'.format(str(validate_fxn)))
        return False

# point is to make improbably that would ever happen to appear inside a pickled image and be mistaken
end_of_message_message = ("---EndOfMessage---"*10) + "\n"   

def send_to_stream(stream, msg): 
    """
    Pickle & send to stdout a message.
    Used primarily to communicate back-and-forth with a separately launched ztv process.
    """
    if isinstance(msg, str):
        msg = (msg,)
    pkl = pickle.dumps(msg)
    stream.write(pkl + '\n' + end_of_message_message)
    stream.flush()

class UnexpectedEndOfStream(Exception): pass

class StreamListenerTimeOut(Exception): pass

def _accumulate_to_queue(stream, queue):
    while True:
        line = stream.readline()
        if line:
            queue.put(line)
        else:
# TODO: rather than return, should really raise the Error, but then code elsewhere needs to be catching for it
#             raise UnexpectedEndOfStream
            return

class StreamListener():
    def __init__(self, stream):
        """
        stream: e.g, stdin/stdout
        """
        self.stream = stream
        self.queue = Queue()
        self.thread = Thread(target=_accumulate_to_queue, args=(self.stream,self.queue))
        self.thread.daemon = True
        self.thread.start() 

    def read_pickled_message(self, timeout=None):
        try:
            block = timeout is not None
            msg = ""
            while not msg.endswith('\n' + end_of_message_message):
                msg += self.queue.get(block=block, timeout=timeout)
            return pickle.loads(msg.replace('\n' + end_of_message_message, ''))
        except Empty:
            raise StreamListenerTimeOut


def listen_to_pipe(pipe, timeout=None):
    """
    Will listen on pipe until has seen end_of_message_message, then strip the 
    end_of_message_message and return the unpickled version of the preceding message
    
    If timeout is None, then will just do a traditional blocking call of readline() 
                        (i.e. will *never* return if a newline never comes in)
    If timeout is not None, then if length of output hasn't changed in timeout seconds,
                        then raise a timeout exception
    """
    in_str = ""
    if timeout is None:
        while not in_str.endswith('\n' + end_of_message_message):
            in_str += pipe.readline()
    else:
        pass
    return pickle.loads(in_str.replace('\n' + end_of_message_message, ''))


from astropy_helpers.git_helpers import get_git_devstr

__all__ = [
    "__title__", "__summary__", "__uri__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
]

__title__ = "ztv"
__summary__ = "Simple python image viewer, largely intended for astronomical applications"
__uri__ = "https://github.com/henryroe/ztv"

# VERSION should be PEP386 compatible (http://www.python.org/dev/peps/pep-0386)
__version__ = "0.1.2dev1"

# Indicates if this version is a release version
RELEASE = 'dev' not in __version__

if not RELEASE:
    __version__ += get_git_devstr(False)

__author__ = "Henry Roe"
__email__ = "hroe@hroe.me"

__license__ = "MIT License"
__copyright__ = "2015 %s" % __author__
